/*
 * File: word.c
 * Date: 2011-11-01
 *
 * Evsiz (evsiz@this-voice.org)
 * Rho released under the Creative Commons NonCommercial ShareAlike License 3.0
 * See http://creativecommons.org for details
 *
 * Rho is a fork of the Linux port of MicroEMACS developed by Humberto
 * Ortiz-Zuazaga and Jean-Pierre Demailly. Their work forked the MicroEMACS
 * of Daniel M. Lawrence which was in turn a fork of uemacs by Dave G.
 * Conroy. MicroEmacs was copyrighted 1993 by Mr. Lawrence with a
 * stipulation that it could not be used commercially. Uemacs was released
 * by Mr. Conroy into the public domain in 1985.
 *
 */


/***

The routines in this file implement commands that work a word at a time.  

***/

#include	<ctype.h>
#include	<stdio.h>
#include	"rho.h"
#include	"word.h"
#include	"cursor.h"
#include	"killring.h"
#include	"line.h"
#include	"macros.h"
#include	"messages.h"
#include	"region.h"
#include	"utils.h"




/*++

Word wrap on n-spaces. Back-over whatever precedes the point on the
current line and stop on the first word-break or the beginning of the
line. If we reach the beginning of the line, jump back to the end of the
word and start a new line.  Otherwise, break the line at the word-break,
eat it, and jump back to the end of the word. Make sure we force the
display back to the left edge of the current window

++*/
int wrap_word(f, n)
int f, n;

{
    int cnt;                    /* size of word wrapped to next line */
    int c;                      /* charector temporary */

    f = n = 0;                  /* splint */

    /* backup from the <NL> 1 char */
    if (!previous_character(FALSE, 1))
        return (FALSE);

    /* back up until we aren't in a word,
       make sure there is a break in the line */
    cnt = 0;
    while (((c = lgetc(curwp->w_dotp, curwp->w_doto)) != ' ')
           && (c != '\t')) {
        cnt++;
        if (!previous_character(FALSE, 1))
            return (FALSE);
        /* if we make it to the beginning, start a new line */
        if (curwp->w_doto == 0) {
            end_of_line();
            return (insertnewline());
        }
    }

    /* delete the forward white space */
    if (!delete_forward(0, 1))
        return (FALSE);

    /* put in a end of line */
    if (!insertnewline())
        return (FALSE);

    /* and past the first word */
    while (cnt-- > 0) {
        if (next_character(FALSE, 1) == FALSE)
            return (FALSE);
    }

    /* make sure the display is not horizontally scrolled */
    if (curwp->w_fcol != 0) {
        curwp->w_fcol = 0;
        curwp->w_flag |= WFHARD | WFMOVE | WFMODE;
    }

    return (TRUE);
}



/*++

Move the cursor backward by "n" words. All of the details of motion are
performed by the "previous_character" and "next_character" routines. Error
if you try to move beyond the buffers.

++*/
int previous_word(f, n)
int f, n;

{
    if (n < 0)
        return (next_word(f, -n));
    if (previous_character(FALSE, 1) == FALSE)
        return (FALSE);
    while (n--) {
        while (inword() == FALSE) {
            if (previous_character(FALSE, 1) == FALSE)
                return (FALSE);
        }
        while (inword() != FALSE) {
            if (previous_character(FALSE, 1) == FALSE)
                return (FALSE);
        }
    }
    return (next_character(FALSE, 1));
}



/*++

Move the cursor forward by the specified number of words. All of the
motion is done by "next_character". Error if you try and move beyond
the buffer's end.

++*/
int next_word(f, n)
int f, n;

{
    if (n < 0)
        return (previous_word(f, -n));
    while (n--) {
        /* scan through the current word */
        while (inword() == TRUE) {
            if (next_character(FALSE, 1) == FALSE)
                return (FALSE);
        }

        /* scan through the intervening white space */
        while (inword() == FALSE) {
            if (next_character(FALSE, 1) == FALSE)
                return (FALSE);
        }
    }
    return (TRUE);
}



/*++

Move forward to the end of the nth next word. Error if you move past
the end of the buffer.

*/
int end_of_word(f, n)
int f, n;

{
    if (n < 0)
        return (previous_word(f, -n));
    while (n--) {
        /* scan through the intervening white space */
        while (inword() == FALSE) {
            if (next_character(FALSE, 1) == FALSE)
                return (FALSE);
        }

        /* scan through the current word */
        while (inword() == TRUE) {
            if (next_character(FALSE, 1) == FALSE)
                return (FALSE);
        }
    }
    return (TRUE);
}



/*++

Move the cursor forward by the specified number of words. As you move,
convert any characters to upper case. Error if you try and move beyond
the end of the buffer. Bound to "M-U".

++*/
int uppercase_word(f, n)
int f, n;

{
    int c;

    f = 0;                      /* splint */

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */
    if (n < 0)
        return (FALSE);
    while (n--) {
        while (inword() == FALSE) {
            if (next_character(FALSE, 1) == FALSE)
                return (FALSE);
        }
        while (inword() != FALSE) {
            c = lgetc(curwp->w_dotp, curwp->w_doto);
            if (islower(c)) {
                c = toupper(c);
                lputc(curwp->w_dotp, curwp->w_doto, c);
                change_line(WFHARD);
            }
            if (next_character(FALSE, 1) == FALSE)
                return (FALSE);
        }
    }
    return (TRUE);
}


/*++

Move the cursor forward by the specified number of words. As you move
convert characters to lower case. Error if you try and move over the
end of the buffer. Bound to "M-L".

++*/
int lowercase_word(f, n)
int f, n;

{
    int c;

    f = 0;                      /* splint */

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */
    if (n < 0)
        return (FALSE);
    while (n--) {
        while (inword() == FALSE) {
            if (next_character(FALSE, 1) == FALSE)
                return (FALSE);
        }
        while (inword() != FALSE) {
            c = lgetc(curwp->w_dotp, curwp->w_doto);
            if (isupper(c)) {
                c = tolower(c);
                lputc(curwp->w_dotp, curwp->w_doto, c);
                change_line(WFHARD);
            }
            if (next_character(FALSE, 1) == FALSE)
                return (FALSE);
        }
    }
    return (TRUE);
}


/*++

Move the cursor forward by the specified number of words. As you move
convert the first character of the word to upper case, and subsequent
characters to lower case. Error if you try and move past the end of the
buffer. Bound to "M-C".

++*/
int capitalize_word(f, n)
int f, n;

{
    int c;

    f = 0;                      /* splint */

    if (curbp->b_mode & MDVIEW) /* don't allow this command if  */
        return (err_readonly());        /* we are in read only mode     */
    if (n < 0)
        return (FALSE);
    while (n--) {
        while (inword() == FALSE) {
            if (next_character(FALSE, 1) == FALSE)
                return (FALSE);
        }
        if (inword() != FALSE) {
            c = lgetc(curwp->w_dotp, curwp->w_doto);
            if (islower(c)) {
                c = toupper(c);
                lputc(curwp->w_dotp, curwp->w_doto, c);
                change_line(WFHARD);
            }
            if (next_character(FALSE, 1) == FALSE)
                return (FALSE);
            while (inword() != FALSE) {
                c = lgetc(curwp->w_dotp, curwp->w_doto);
                if (isupper(c)) {
                    c = tolower(c);
                    lputc(curwp->w_dotp, curwp->w_doto, c);
                    change_line(WFHARD);
                }
                if (next_character(FALSE, 1) == FALSE)
                    return (FALSE);
            }
        }
    }
    return (TRUE);
}



/*++

Kill forward by "n" words. Remember the location of dot. Move forward by
the right number of words. Put dot back where it was and issue the kill
command for the right number of characters. With a zero argument, just
kill one word and no whitespace. Bound to "M-D".

++*/
int delete_next_word(f, n)
int f, n;

{
    LINE *dotp;                 /* original cursor line */
    int doto;                   /*      and row */
    int c;                      /* temp char */
    long size;                  /* # of chars to delete */

    f = 0;                      /* splint */

    /* don't allow this command if we are in read only mode */
    if (curbp->b_mode & MDVIEW)
        return (err_readonly());

    /* ignore the command if there is a negative argument */
    if (n < 0)
        return (FALSE);

    /* Clear the kill buffer if last command wasn't a kill */
    if ((lastflag & CFKILL) == 0)
        next_kill();
    thisflag |= CFKILL;         /* this command is a kill */

    /* save the current cursor position */
    dotp = curwp->w_dotp;
    doto = curwp->w_doto;

    /* figure out how many characters to give the axe */
    size = 0;

    /* get us into a word.... */
    while (inword() == FALSE) {
        if (next_character(FALSE, 1) == FALSE)
            return (FALSE);
        ++size;
    }

    if (n == 0) {
        /* skip one word, no whitespace! */
        while (inword() == TRUE) {
            if (next_character(FALSE, 1) == FALSE)
                return (FALSE);
            ++size;
        }
    } else {
        /* skip n words.... */
        while (n--) {

            /* if we are at EOL; skip to the beginning of the next */
            while (curwp->w_doto == lused(curwp->w_dotp)) {
                if (next_character(FALSE, 1) == FALSE)
                    return (FALSE);
                ++size;
            }

            /* move forward till we are at the end of the word */
            while (inword() == TRUE) {
                if (next_character(FALSE, 1) == FALSE)
                    return (FALSE);
                ++size;
            }

            /* if there are more words, skip the interword stuff */
            if (n != 0)
                while (inword() == FALSE) {
                    if (next_character(FALSE, 1) == FALSE)
                        return (FALSE);
                    ++size;
                }
        }

        /* skip whitespace and newlines */
        while ((curwp->w_doto == lused(curwp->w_dotp)) ||
               ((c = lgetc(curwp->w_dotp, curwp->w_doto)) == ' ') ||
               (c == '\t')) {
            if (next_character(FALSE, 1) == FALSE)
                break;
            ++size;
        }
    }

    /* restore the original position and delete the words */
    curwp->w_dotp = dotp;
    curwp->w_doto = doto;
    return (deletenchars(size, TRUE));
}



/*++

Kill backwards by "n" words. Move backwards by the desired number of words,
counting the characters. When dot is finally moved to its resting place,
fire off the kill command. Bound to "M-Rubout" and to "M-Backspace".

++*/
int delete_previous_word(f, n)
int f, n;

{
    long size;

    f = 0;                      /* splint */

    /* don't allow this command if we are in read only mode */
    if (curbp->b_mode & MDVIEW)
        return (err_readonly());

    /* ignore the command if there is a nonpositive argument */
    if (n <= 0)
        return (FALSE);

    /* Clear the kill buffer if last command wasn't a kill */
    if ((lastflag & CFKILL) == 0)
        next_kill();
    thisflag |= CFKILL;         /* this command is a kill */

    if (previous_character(FALSE, 1) == FALSE)
        return (FALSE);
    size = 0;
    while (n--) {
        while (inword() == FALSE) {
            if (previous_character(FALSE, 1) == FALSE)
                return (FALSE);
            ++size;
        }
        while (inword() != FALSE) {
            ++size;
            if (previous_character(FALSE, 1) == FALSE)
                goto bckdel;
        }
    }
    if (next_character(FALSE, 1) == FALSE)
        return (FALSE);
  bckdel:if (next_character(FALSE, size) == FALSE)
        return (FALSE);
    return (deletenchars(-size, TRUE));
}



/*++

Return TRUE if the character at dot is a character that is considered to be
part of a word. The default word character list is hard coded. If $wchars
has been set by the user, use that instead.
Used by most fns in this file.

++*/
int inword()
{
    int c;

    /* the end of a line is never in a word */
    if (curwp->w_doto == lused(curwp->w_dotp))
        return (FALSE);

    /* grab the word to check */
    c = lgetc(curwp->w_dotp, curwp->w_doto);

    /* if we are using the table.... */
    if (wlflag)
        return (wordlist[c]);

    /* else use the default hard coded check */
    if (isalpha(c))
        return (TRUE);
    if (c >= '0' && c <= '9')
        return (TRUE);
    if (c == '_')
        return (TRUE);
    return (FALSE);
}



/*++

Word and char count

++*/
int wordcount(f, n)
int f, n;                       /* ignored numeric arguments */

{
    LINE *lp;                   /* current line to scan */
    int offset;                 /* current char to scan */
    long size;                  /* size of region left to count */
    int ch;                     /* current character to scan */
    int wordflag;               /* are we in a word now? */
    int lastword;               /* were we just in a word? */
    long nwords;                /* total # of words */
    long nchars;                /* total number of chars */
    int status;                 /* status return code */
    REGION region;              /* region to look at */

    f = n = 0;                  /* splint */

    /* make sure we have a region to count */
    if ((status = getregion(&region)) != TRUE)
        return (status);
    lp = region.r_linep;
    offset = region.r_offset;
    size = region.r_size;

    /* count up things */
    lastword = FALSE;
    nwords = 0L;
    while (size--) {

        /* get the current character */
        if (offset == lused(lp)) {      /* end of line */
            ch = '\r';
            lp = lforw(lp);
            offset = 0;
        } else {
            ch = lgetc(lp, offset);
            ++offset;
        }

        /* and tabulate it */
        wordflag = ((ch >= 'a' && ch <= 'z') ||
                    (ch >= 'A' && ch <= 'Z') || (ch >= '0' && ch <= '9'));
        if (wordflag == TRUE && lastword == FALSE)
            ++nwords;
        lastword = wordflag;
        ++nchars;
    }

    mlwrite(TEXT100,            /*"Words: %D  Chars: %D" */
            nwords, nchars);
    return (TRUE);
}
